#!/usr/bin/env python3
"""
An Interesting Consequence of Mathematics
by Javantea
April 8, 2015

What AES key produces the same first four bytes of keystream as another key?
It should only take testing a few thousand keys to find a single instance of 
this, so let's do it. It's super effective.

"""
import Crypto.Cipher.AES
import sys
import struct
import random
import Crypto.Util.number
import Crypto.Util.randpool

def collideCiphertext(ciphertexts, keys=0x100000, start = 0, length=128):
	if length != 128 and length != 256:
		raise ValueError("Length must be 128 or 256.")
	#end if
	output = []
	i = start
	while len(output) < keys:
		if length == 128:
			# 128-bit key.
			key = struct.pack('>QQ', i >> 64, i & 0xffffffffffffffff)
		elif length == 256:
			# 256-bit key.
			key = struct.pack('>QQQQ', i >> 192, 
				(i >> 128) & 0xffffffffffffffff, 
				(i >> 64) & 0xffffffffffffffff, 
				i & 0xffffffffffffffff)
		#end if
		ciphertext = Crypto.Cipher.AES.new(key).encrypt(b'GET / HTTP/1.1\r\n')
		if ciphertext in ciphertexts:
			output.append(key)
		i += 1
		if i & 0xffffff == 0: print(i)
	#loop
	return output, i
#end def collideCiphertext(ciphertexts, keys=0x100000, start = 0, length = 128)

def collideRandomCiphertext(ciphertexts, keys=0x100000, length=128):
	if length != 128 and length != 256:
		raise ValueError("Length must be 128 or 256.")
	#end if
	rand = Crypto.Util.randpool.RandomPool()
	output = []
	i = 0
	while len(output) < keys:
		key = rand.get_bytes(length>>3)
		ciphertext = Crypto.Cipher.AES.new(key).encrypt(b'GET / HTTP/1.1\r\n')
		if ciphertext[:4] in ciphertexts:
			#print('found', i, key, ciphertext)
			output.append(key)
		i += 1
		if i & 0xffffff == 0: print(i)
	#loop
	return output, i
#end def collideCiphertext(ciphertexts, keys=0x100000, start = 0, length = 128)

def getCiphertexts(start = 0, ciphertexts=0x100000, length=128):
	if length != 128 and length != 256:
		raise ValueError("Length must be 128 or 256.")
	#end if
	output = []
	i = start
	for v in range(ciphertexts):
		if length == 128:
			# 128-bit key.
			key = struct.pack('>QQ', i >> 64, i & 0xffffffffffffffff)
		elif length == 256:
			# 256-bit key.
			key = struct.pack('>QQQQ', i >> 192, 
				(i >> 128) & 0xffffffffffffffff, 
				(i >> 64) & 0xffffffffffffffff, 
				i & 0xffffffffffffffff)
		#end if
		ciphertext = Crypto.Cipher.AES.new(key).encrypt(b'GET / HTTP/1.1\r\n')
		# Only take the first four bytes.
		output.append(ciphertext[:4])
		i += 1
		if i & 0xffffff == 0: print(i)
	#loop
	return output
#end def getKeyStreams(start = 0, ciphertexts=0x100000, length=128)

def main():
	start = 0x100000
	randomize = False
	keys_we_want = 1024
	if '-r' in sys.argv:
		randomize = True   
		sys.argv.pop(sys.argv.index('-r'))
	#end if
	if len(sys.argv) > 1:
		start = int(sys.argv[1])
		if len(sys.argv) > 2:
			keys_we_want = int(sys.argv[2])
		#end if
	#end if
	print("Starting at:", start)
	if start < 0x100000:
		print("This isn't going to work. start should be greater than", 0x100000)
	#end if
	
	tons_of_ciphertexts = getCiphertexts(0, 0x100000)
	tons_of_ciphertexts_set = set(tons_of_ciphertexts)
	print('len(tons_of_ciphertexts_set):', len(tons_of_ciphertexts_set), '%3.3f' % (len(tons_of_ciphertexts_set) / 0x100000))
	if randomize:
		keys_out, end = collideRandomCiphertext(tons_of_ciphertexts_set, keys_we_want)
	else:
		keys_out, end = collideCiphertext(tons_of_ciphertexts_set, keys_we_want, start)
	#end if
	for key in keys_out:
		ciphertext = Crypto.Cipher.AES.new(key).encrypt(b'GET / HTTP/1.1\r\n')
		print(key, ciphertext, tons_of_ciphertexts.index(ciphertext[:4]))
	#next key
	print(len(keys_out), "keys found in", end-start, "attempts")
#end def main()

if __name__ == '__main__':
	main()
#end if


